import functools
from typing import Any, Iterable, Optional, Sequence

from mediator.common.factory.cascade import HandlerFactoryCascade
from mediator.common.factory.policies import PolicyType
from mediator.common.operators import OperatorDef
from mediator.common.registry.base import HandlerEntry, HandlerStore
from mediator.common.registry.stores import OperatorHandlerStore


class AbstractHandlerRegistry(HandlerStore):
    def __init__(
        self,
        policies: Optional[Sequence[PolicyType]] = None,
        cascade: Optional[HandlerFactoryCascade] = None,
    ):
        self._cascade = self._provide_cascade(policies=policies, cascade=cascade)

    @classmethod
    def _provide_cascade(
        cls,
        policies: Optional[Sequence[PolicyType]],
        cascade: Optional[HandlerFactoryCascade],
    ):
        if cascade is not None:
            return cascade
        if policies is not None:
            return HandlerFactoryCascade(policies=policies)
        raise TypeError("policies and cascade are not defined")

    def register(
        self,
        obj: Optional[Any] = None,
        *,
        policies: Optional[Sequence[PolicyType]] = None,
        operators: Sequence[OperatorDef] = (),
    ):
        cascade = self._policies_cascade(policies)
        if obj is None:
            process = functools.partial(
                self._process_obj, cascade=cascade, operators=operators
            )

            def wrap(fn: Any):
                process(fn)
                return fn

            return wrap

        else:
            self._process_obj(obj, cascade=cascade, operators=operators)
            return obj

    def _policies_cascade(
        self, policies: Optional[Sequence[PolicyType]]
    ) -> HandlerFactoryCascade:
        if policies is None:
            return self._cascade
        else:
            return self._cascade.replace(policies)

    def _process_obj(
        self,
        obj: Any,
        *,
        cascade: HandlerFactoryCascade,
        operators: Sequence[OperatorDef],
    ):
        handler = cascade(obj)
        self.add(HandlerEntry(handler=handler, operators=operators))

    def add(self, entry: HandlerEntry):
        raise NotImplementedError

    def include(self, entries: Iterable[HandlerEntry]):
        raise NotImplementedError

    def __iter__(self) -> Iterable[HandlerEntry]:
        raise NotImplementedError


class HandlerRegistry(AbstractHandlerRegistry):
    def __init__(
        self,
        store: HandlerStore,
        policies: Optional[Sequence[PolicyType]] = None,
        cascade: Optional[HandlerFactoryCascade] = None,
        operators: Sequence[OperatorDef] = (),
    ):
        AbstractHandlerRegistry.__init__(self, policies=policies, cascade=cascade)
        self._store = OperatorHandlerStore.wrap(nested=store, operators=operators)

    def add(self, entry: HandlerEntry):
        self._store.add(entry)

    def include(self, entries: Iterable[HandlerEntry]):
        self._store.include(entries)

    def __iter__(self) -> Iterable[HandlerEntry]:
        return iter(self._store)



from typing import Generic, Optional, TypeVar

T = TypeVar("T")


class AsyncNullContextManager(Generic[T]):
    def __init__(self, return_value: Optional[T] = None):
        self.return_value = return_value

    async def __aenter__(self):
        return self.return_value

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass



from mediator.common.registry.base import (
    CollisionHandlerStoreError,
    HandlerEntry,
    HandlerStore,
    HandlerStoreError,
    InspectionHandlerStoreError,
    LookupHandlerStoreError,
)
from mediator.common.registry.registry import AbstractHandlerRegistry, HandlerRegistry
from mediator.common.registry.stores import (
    CollectionHandlerStore,
    MapHandlerStore,
    OperatorHandlerStore,
)

__all__ = [
    # base
    "CollisionHandlerStoreError",
    "HandlerEntry",
    "HandlerStore",
    "HandlerStoreError",
    "InspectionHandlerStoreError",
    "LookupHandlerStoreError",
    # registry
    "AbstractHandlerRegistry",
    "HandlerRegistry",
    # stores
    "CollectionHandlerStore",
    "MapHandlerStore",
    "OperatorHandlerStore",
]



from mediator.event.aggregate import (
    ConfigEventAggregateError,
    EventAggregate,
    EventAggregateError,
)
from mediator.event.base import EventPublish, EventPublisher, EventSubscriber
from mediator.event.local import LocalEventBus
from mediator.event.registry import EventHandlerRegistry

__all__ = [
    "ConfigEventAggregateError",
    "EventAggregate",
    "EventAggregateError",
    "EventPublish",
    "EventPublisher",
    "EventSubscriber",
    "LocalEventBus",
    "EventHandlerRegistry",
]



from typing import Any, Iterable, Sequence

from mediator.common.handler import Handler


class HandlerFactoryError(Exception):
    pass


class IncompatibleHandlerFactoryError(TypeError, HandlerFactoryError):
    pass


class HandlerFactory:
    def create(self, obj: Any) -> Handler:
        raise NotImplementedError


class HandlerFactoryMapper:
    def map_all(self, policies: Iterable[Any]) -> Sequence[HandlerFactory]:
        return [self.map(policy) for policy in policies]

    def map(self, policy: Any) -> HandlerFactory:
        raise NotImplementedError



from typing import Any, AsyncContextManager, Iterable

from mediator.common.registry import HandlerEntry, HandlerStore
from mediator.utils.context import AsyncNullContextManager


class EventPublish:
    async def publish(self, obj: Any, **kwargs):
        raise NotImplementedError


class EventPublisher(EventPublish):
    # noinspection PyMethodMayBeStatic
    def transaction(self) -> AsyncContextManager[EventPublish]:
        return AsyncNullContextManager(self)

    async def publish(self, obj: Any, **kwargs):
        raise NotImplementedError


class EventSubscriber(HandlerStore):
    def add(self, entry: HandlerEntry):
        raise NotImplementedError

    def include(self, entries: Iterable[HandlerEntry]):
        raise NotImplementedError

    def __iter__(self) -> Iterable[HandlerEntry]:
        raise NotImplementedError



from typing import Any, Iterable, Optional, Sequence, Tuple

from mediator.common.factory.base import (
    HandlerFactoryError,
    IncompatibleHandlerFactoryError,
)
from mediator.common.factory.mappers import (
    DefaultHandlerFactoryMapper,
    HandlerFactory,
    HandlerFactoryMapper,
)
from mediator.common.factory.policies import PolicyType
from mediator.common.handler import Handler


class HandlerFactoryCascadeError(Exception):
    pass


class ConfigHandlerFactoryCascadeError(ValueError, HandlerFactoryCascadeError):
    pass


class InspectionHandlerFactoryCascadeError(HandlerFactoryCascadeError):
    def __init__(self, *args, factory: HandlerFactory):
        super().__init__(*args)
        self.factory = factory


class IncompatibleHandlerFactoryCascadeError(HandlerFactoryCascadeError):
    def __init__(
        self,
        *args,
        factory_errors: Sequence[
            Tuple[HandlerFactory, IncompatibleHandlerFactoryError]
        ],
    ):
        super().__init__(*args)
        self.factory_errors = factory_errors


class HandlerFactoryCascade:
    def __init__(
        self,
        policies: Iterable[PolicyType],
        mapper: Optional[HandlerFactoryMapper] = None,
    ):
        policies = [*policies]
        if not policies:
            raise ConfigHandlerFactoryCascadeError("No policies provided")
        if mapper is None:
            mapper = DefaultHandlerFactoryMapper()
        self._mapper = mapper
        self._factories = mapper.map_all(policies)

    def __call__(self, obj: Any) -> Handler:
        factory_errors = []
        for factory in self._factories:
            try:
                return factory.create(obj)
            except IncompatibleHandlerFactoryError as e:
                factory_errors.append((factory, e))
                continue
            except HandlerFactoryError as error:
                raise InspectionHandlerFactoryCascadeError(
                    f"Handler factory {factory!r} error for object {obj!r}: {error}",
                    factory=factory,
                ) from error
        raise IncompatibleHandlerFactoryCascadeError(
            f"Object {obj!r} is not compatible",
            factory_errors=factory_errors,
        )

    def replace(self, policies: Iterable[PolicyType]):
        return HandlerFactoryCascade(policies=policies, mapper=self._mapper)



from mediator.common.factory.base import (
    HandlerFactory,
    HandlerFactoryError,
    HandlerFactoryMapper,
    IncompatibleHandlerFactoryError,
)
from mediator.common.factory.cascade import (
    ConfigHandlerFactoryCascadeError,
    HandlerFactoryCascade,
    HandlerFactoryCascadeError,
    IncompatibleHandlerFactoryCascadeError,
    InspectionHandlerFactoryCascadeError,
)
from mediator.common.factory.factories import (
    CallableHandlerFactory,
    MethodHandlerFactory,
)
from mediator.common.factory.mappers import (
    DefaultHandlerFactoryMapper,
    TypeHandlerFactoryMapper,
)
from mediator.common.factory.policies import (
    CallableHandlerPolicy,
    MappablePolicy,
    MethodHandlerPolicy,
    PolicyType,
)

__all__ = [
    # base
    "HandlerFactory",
    "HandlerFactoryError",
    "HandlerFactoryMapper",
    "IncompatibleHandlerFactoryError",
    # cascade
    "ConfigHandlerFactoryCascadeError",
    "HandlerFactoryCascade",
    "HandlerFactoryCascadeError",
    "IncompatibleHandlerFactoryCascadeError",
    "InspectionHandlerFactoryCascadeError",
    # factories
    "CallableHandlerFactory",
    "MethodHandlerFactory",
    # mappers
    "DefaultHandlerFactoryMapper",
    "TypeHandlerFactoryMapper",
    # policy
    "CallableHandlerPolicy",
    "MappablePolicy",
    "MethodHandlerPolicy",
    "PolicyType",
]



from typing import Any, Callable, Mapping, Optional, Type

from mediator.common.factory.base import HandlerFactory, HandlerFactoryMapper
from mediator.common.factory.factories import (
    CallableHandlerFactory,
    MethodHandlerFactory,
)
from mediator.common.factory.policies import CallableHandlerPolicy, MethodHandlerPolicy


class TypeHandlerFactoryMapper(HandlerFactoryMapper):
    def __init__(self, mapping: Mapping[Type, Callable[[Any], HandlerFactory]]):
        self._mapping = mapping

    def map(self, policy: Any) -> HandlerFactory:
        policy_type = type(policy)
        handler_factory = self._mapping.get(policy_type)
        if handler_factory is None:
            raise TypeError(f"Not supported policy type {policy_type}")
        return handler_factory(policy)


class DefaultHandlerFactoryMapper(TypeHandlerFactoryMapper):
    def __init__(
        self,
        extra_mapping: Optional[Mapping[Type, Callable[[Any], HandlerFactory]]] = None,
    ):
        if extra_mapping is None:
            extra_mapping = {}
        super().__init__(
            {
                **self.default_mapping(),
                **extra_mapping,
            }
        )

    @classmethod
    def default_mapping(cls):
        return {
            CallableHandlerPolicy: CallableHandlerFactory,
            MethodHandlerPolicy: MethodHandlerFactory,
        }



from typing import Any, Dict, Hashable, Iterable, Optional, Sequence

from mediator.common.factory import (
    CallableHandlerPolicy,
    HandlerFactoryCascade,
    PolicyType,
)
from mediator.common.operators import OperatorDef
from mediator.common.registry import (
    HandlerEntry,
    HandlerRegistry,
    LookupHandlerStoreError,
    MapHandlerStore,
)
from mediator.common.types import ActionCallType, ActionResult, ActionSubject
from mediator.request.base import RequestExecutor


class _RequestExecutorHandlerStore(MapHandlerStore):
    _calls: Dict[Hashable, ActionCallType]

    def __init__(self):
        super().__init__()
        self._calls = {}

    def add(self, entry: HandlerEntry):
        super().add(entry)
        self._map_call(entry)

    def include(self, entries: Iterable[HandlerEntry]):
        super().include(entries)
        for entry in entries:
            self._map_call(entry)

    def _map_call(self, entry: HandlerEntry):
        self._calls[entry.key] = entry.handler_pipeline()

    async def __call__(self, action: ActionSubject) -> ActionResult:
        key = action.key
        call = self._calls.get(key)
        if call is None:
            raise LookupHandlerStoreError(f"Handler not defined for key {key}")
        return await call(action)


class LocalRequestBus(HandlerRegistry, RequestExecutor):
    def __init__(
        self,
        policies: Optional[Sequence[PolicyType]] = None,
        cascade: Optional[HandlerFactoryCascade] = None,
        operators: Sequence[OperatorDef] = (),
    ):
        executor_store = _RequestExecutorHandlerStore()
        HandlerRegistry.__init__(
            self,
            store=executor_store,
            policies=policies or [CallableHandlerPolicy()],
            cascade=cascade,
            operators=operators,
        )
        self._executor = executor_store

    async def execute(self, obj: Any, **kwargs):
        result = await self._executor(ActionSubject(subject=obj, inject=kwargs))
        assert isinstance(
            result, ActionResult
        ), "operator or handler should provide `ActionResult` type object"
        return result.result



from dataclasses import dataclass
from typing import Any, Awaitable, Callable, Dict, Hashable


@dataclass
class ActionSubject:
    subject: Any
    inject: Dict[str, Any]

    @property
    def key(self) -> Hashable:
        return type(self.subject)


@dataclass
class ActionResult:
    result: Any


ActionCallType = Callable[[ActionSubject], Awaitable[ActionResult]]



from typing import Any, Hashable

from mediator.common.types import ActionResult, ActionSubject


class HandlerInfo:
    @property
    def key(self) -> Hashable:
        raise NotImplementedError

    @property
    def obj(self) -> Any:
        raise NotImplementedError


class Handler(HandlerInfo):
    @property
    def key(self) -> Hashable:
        raise NotImplementedError

    @property
    def obj(self) -> Any:
        raise NotImplementedError

    async def __call__(self, action: ActionSubject) -> ActionResult:
        raise NotImplementedError



from typing import Any, Awaitable, Callable, Collection, Dict, Hashable, Optional, Tuple

from mediator.common.handler.base import Handler
from mediator.common.types import ActionResult, ActionSubject


class DirectHandler(Handler):
    def __init__(self, handler: Any, fn: Callable[..., Awaitable[Any]], key: Hashable):
        self._handler = handler
        self._fn = fn
        self._key = key

    @property
    def key(self) -> Hashable:
        return self._key

    @property
    def obj(self) -> Any:
        return self._handler

    async def __call__(self, action: ActionSubject) -> ActionResult:
        result = await self._fn(action.subject, **action.inject)
        return ActionResult(result)


class MappedDirectHandler(DirectHandler):
    def __init__(
        self,
        handler: Any,
        fn: Callable[..., Awaitable[Any]],
        key: Hashable,
        subject_name: Optional[str],
        arg_map: Dict[str, str],
        arg_filter: Optional[Collection[str]],
    ):
        super().__init__(handler, fn, key)
        self._arg_map = self._arg_map_factory(arg_map)
        self._arg_filter = self._arg_filter_factory(arg_filter)
        self._args = self._args_factory(subject_name)

    async def __call__(self, action: ActionSubject) -> ActionResult:
        kwargs = self._arg_map(action.inject)
        kwargs = self._arg_filter(kwargs)
        args, kwargs = self._args(action.subject, kwargs)
        result = await self._fn(*args, **kwargs)
        return ActionResult(result)

    @classmethod
    def _arg_map_factory(cls, arg_map: Dict[str, str]):
        if arg_map:

            def _arg_map(kwargs: Dict[str, Any]):
                return {
                    arg_map.get(name, name): value for name, value in kwargs.items()
                }

        else:

            def _arg_map(kwargs: Dict[str, Any]):
                return kwargs

        return _arg_map

    @classmethod
    def _arg_filter_factory(cls, arg_filter: Optional[Collection[str]]):
        if arg_filter is None:

            def _arg_filter(kwargs: Dict[str, Any]):
                return kwargs

        else:
            args = set(arg_filter)

            def _arg_filter(kwargs: Dict[str, Any]):
                return {name: value for name, value in kwargs.items() if name in args}

        return _arg_filter

    @classmethod
    def _args_factory(cls, subject_key: Optional[str]):
        if subject_key:

            def _args(
                subject: Any, kwargs: Dict[str, Any]
            ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
                kwargs[subject_key] = subject
                return (), kwargs

        else:

            def _args(
                subject: Any, kwargs: Dict[str, Any]
            ) -> Tuple[Tuple[Any, ...], Dict[str, Any]]:
                return (subject,), kwargs

        return _args



from typing import Optional, Sequence

from mediator.common.factory import HandlerFactoryCascade, PolicyType
from mediator.common.operators import OperatorDef
from mediator.common.registry import CollectionHandlerStore, HandlerRegistry


class EventHandlerRegistry(HandlerRegistry):
    def __init__(
        self,
        policies: Optional[Sequence[PolicyType]] = None,
        cascade: Optional[HandlerFactoryCascade] = None,
        operators: Sequence[OperatorDef] = (),
    ):
        HandlerRegistry.__init__(
            self,
            store=CollectionHandlerStore(),
            policies=policies,
            cascade=cascade,
            operators=operators,
        )



import asyncio
from collections import defaultdict
from typing import Any, Dict, Hashable, Iterable, List, Optional, Sequence

from mediator.common.factory import (
    CallableHandlerPolicy,
    HandlerFactoryCascade,
    PolicyType,
)
from mediator.common.operators import OperatorDef
from mediator.common.registry import (
    CollectionHandlerStore,
    HandlerEntry,
    HandlerRegistry,
)
from mediator.common.types import ActionCallType, ActionSubject
from mediator.event.base import EventPublisher, EventSubscriber


class _EventSchedulerHandlerStore(CollectionHandlerStore):
    _groups: Dict[Hashable, List[ActionCallType]]

    def __init__(self, sync_mode: bool = False):
        super().__init__()
        self._groups = defaultdict(list)
        self.sync_mode = sync_mode

    def add(self, entry: HandlerEntry):
        super().add(entry)
        self._map_call(entry)

    def include(self, entries: Iterable[HandlerEntry]):
        super().include(entries)
        for entry in entries:
            self._map_call(entry)

    def _map_call(self, entry: HandlerEntry):
        self._groups[entry.key].append(entry.handler_pipeline())

    async def schedule(self, action: ActionSubject):
        key = action.key
        group: List[ActionCallType] = self._groups.get(key, ())
        tasks = [asyncio.create_task(call(action)) for call in group]
        if self.sync_mode:
            await asyncio.wait(tasks)


class LocalEventBus(EventPublisher, HandlerRegistry, EventSubscriber):
    def __init__(
        self,
        policies: Optional[Sequence[PolicyType]] = None,
        cascade: Optional[HandlerFactoryCascade] = None,
        operators: Sequence[OperatorDef] = (),
        sync_mode: bool = False,
    ):
        scheduler_store = _EventSchedulerHandlerStore(sync_mode=sync_mode)
        HandlerRegistry.__init__(
            self,
            store=scheduler_store,
            policies=policies or [CallableHandlerPolicy()],
            cascade=cascade,
            operators=operators,
        )
        self._scheduler = scheduler_store

    async def publish(self, obj: Any, **kwargs):
        await self._scheduler.schedule(ActionSubject(subject=obj, inject=kwargs))



from typing import Dict, Hashable, Iterable, List, Sequence

from mediator.common.operators import OperatorDef
from mediator.common.registry.base import (
    CollisionHandlerStoreError,
    HandlerEntry,
    HandlerStore,
    LookupHandlerStoreError,
)


class CollectionHandlerStore(HandlerStore):
    def __init__(self):
        self._entries: List[HandlerEntry] = []

    def add(self, entry: HandlerEntry):
        self._entries.append(entry)

    def include(self, entries: Iterable[HandlerEntry]):
        self._entries.extend(entries)

    def __iter__(self) -> Iterable[HandlerEntry]:
        return iter(self._entries)


class MapHandlerStore(HandlerStore):
    def __init__(self):
        self._map: Dict[Hashable, HandlerEntry] = {}

    def add(self, entry: HandlerEntry):
        key = entry.key
        if key in self._map:
            raise CollisionHandlerStoreError(
                f"Handler entry with key {key} already exists"
            )
        self._map[entry.key] = entry

    def include(self, entries: Iterable[HandlerEntry]):
        for entry in entries:
            self.add(entry)

    def __iter__(self) -> Iterable[HandlerEntry]:
        return iter(self._map.values())

    def get(self, key: Hashable) -> HandlerEntry:
        if key not in self._map:
            raise LookupHandlerStoreError(f"Handler entry for key {key} not found")
        return self._map[key]


class OperatorHandlerStore(HandlerStore):
    def __init__(self, nested: HandlerStore, operators: Sequence[OperatorDef]):
        self._nested = nested
        self._operators = tuple(operators)

    def add(self, entry: HandlerEntry):
        self._nested.add(entry.stack(self._operators))

    def include(self, entries: Iterable[HandlerEntry]):
        for entry in entries:
            self.add(entry)

    def __iter__(self) -> Iterable[HandlerEntry]:
        return iter(self._nested)

    @classmethod
    def wrap(cls, nested: HandlerStore, operators: Sequence[OperatorDef]):
        if operators:
            return cls(nested=nested, operators=operators)
        return nested



from typing import Any, Dict, Optional, Sequence, Tuple

from mediator.common.factory.base import (
    HandlerFactoryError,
    IncompatibleHandlerFactoryError,
)
from mediator.common.handler import Handler, MappedDirectHandler
from mediator.utils.inspection import CallableArg, CallableDetails, CallableInspector


class CallableObjDetails:
    def __call__(self, obj: Any) -> CallableDetails:
        try:
            return CallableInspector.inspect(obj)
        except TypeError as e:
            raise IncompatibleHandlerFactoryError(f"{obj!r}: {e}")


class CallableAttributeDetails:
    name: str
    subtype_of: Tuple[Any, ...]

    def __init__(self, name: str, owner_subtype_of: Sequence[Any] = ()):
        self.name = name
        self.subtype_of = tuple(owner_subtype_of)

    def __call__(self, obj: Any) -> CallableDetails:
        if self.subtype_of:
            if not isinstance(obj, self.subtype_of):
                raise IncompatibleHandlerFactoryError(
                    f"Object {obj!r} is not type of {self.subtype_of}"
                )

        try:
            attr = getattr(obj, self.name)
        except AttributeError:
            raise IncompatibleHandlerFactoryError(
                f"Object {obj!r} has no attribute named {self.name}"
            )

        try:
            return CallableInspector.inspect(attr)
        except TypeError as e:
            raise IncompatibleHandlerFactoryError(f"{obj!r} attribute {self.name}: {e}")


class HandlerSubjectArgGet:
    def __init__(self, name: Optional[str] = None):
        self.name = name

    def __call__(self, details: CallableDetails) -> CallableArg:
        if not details.args:
            raise IncompatibleHandlerFactoryError(
                f"Callable {details.obj!r} has no explicit argument"
            )
        arg = self._find(details)
        self._check_type(details, arg)
        return arg

    def _find(self, details: CallableDetails) -> CallableArg:
        if self.name:
            return self._find_by_name(details, self.name)
        else:
            return self._get_first(details)

    @staticmethod
    def _find_by_name(details: CallableDetails, name: str):
        arg = details.arg_by_name(name)
        if not arg:
            raise IncompatibleHandlerFactoryError(
                f"Callable {details.obj!r} has no argument named {name}"
            )
        return arg

    @staticmethod
    def _get_first(details: CallableDetails):
        return details.args[0]

    @staticmethod
    def _check_type(details: CallableDetails, arg: CallableArg):
        if arg.type is None:
            raise HandlerFactoryError(
                f"Callable {details.obj!r} argument {arg.name} has no type annotation"
            )
        if not isinstance(arg.type, type):
            raise HandlerFactoryError(
                f"Callable {details.obj!r} argument {arg.name}"
                f" type annotation is not a type"
            )


class DirectHandlerCreate:
    def __init__(
        self, subject_as_keyword: bool, arg_map: Dict[str, str], arg_strict: bool
    ):
        self.subject_as_keyword = subject_as_keyword
        self.arg_map = arg_map
        self.arg_strict = arg_strict

    def __call__(self, details: CallableDetails, arg: CallableArg, obj: Any) -> Handler:
        if not details.is_async:
            raise HandlerFactoryError(f"Object {details.obj!r} is not async callable")
        if self.subject_as_keyword or not arg.is_positional:
            subject_name = arg.name
        else:
            subject_name = None
        if self.arg_strict or details.has_kwargs:
            arg_filter = None
        else:
            arg_filter = {arg.name for arg in details.args}
        return MappedDirectHandler(
            handler=obj,
            fn=details.obj,
            key=arg.type,
            subject_name=subject_name,
            arg_map=self.arg_map,
            arg_filter=arg_filter,
        )



from dataclasses import dataclass
from typing import Iterable, Sequence

from mediator.common.handler.base import Handler
from mediator.common.operators import OperatorDef, OperatorStack
from mediator.common.types import ActionCallType


@dataclass
class HandlerEntry:
    handler: Handler
    operators: Sequence[OperatorDef] = ()

    @property
    def key(self):
        return self.handler.key

    def handler_pipeline(self) -> ActionCallType:
        return self.pipeline(self.handler)

    def pipeline(self, call: ActionCallType) -> ActionCallType:
        return OperatorStack.build(self.operators, call)

    def stack(self, operators: Sequence[OperatorDef]) -> "HandlerEntry":
        return HandlerEntry(
            handler=self.handler,
            operators=tuple([*operators, *self.operators]),
        )


class HandlerStoreError(Exception):
    pass


class InspectionHandlerStoreError(HandlerStoreError):
    pass


class CollisionHandlerStoreError(KeyError, HandlerStoreError):
    pass


class LookupHandlerStoreError(LookupError, HandlerStoreError):
    pass


class HandlerStore:
    def add(self, entry: HandlerEntry):
        raise NotImplementedError

    def include(self, entries: Iterable[HandlerEntry]):
        raise NotImplementedError

    def __iter__(self) -> Iterable[HandlerEntry]:
        raise NotImplementedError



from mediator.request.base import RequestExecutor
from mediator.request.local import LocalRequestBus
from mediator.request.registry import RequestHandlerRegistry

__all__ = [
    "RequestExecutor",
    "LocalRequestBus",
    "RequestHandlerRegistry",
]



import inspect
from dataclasses import dataclass
from inspect import Parameter
from typing import Any, List, Optional, get_type_hints


@dataclass
class CallableArg:
    name: str
    type: Optional[Any]
    is_positional: bool


@dataclass
class CallableType:
    is_routine: bool
    is_method: bool
    is_async: bool


@dataclass
class CallableIODetails:
    args: List[CallableArg]
    has_kwargs: bool
    return_type: Optional[Any]

    def arg_by_name(self, name: str) -> Optional[CallableArg]:
        for arg in self.args:
            if arg.name == name:
                return arg
        return None


@dataclass
class CallableDetails(CallableType, CallableIODetails):
    obj: Any


class CallableInspector:
    _kinds = (
        Parameter.POSITIONAL_ONLY,
        Parameter.POSITIONAL_OR_KEYWORD,
        Parameter.KEYWORD_ONLY,
    )
    _positional_kinds = (Parameter.POSITIONAL_ONLY, Parameter.POSITIONAL_OR_KEYWORD)

    @classmethod
    def inspect(cls, fn: Any) -> CallableDetails:
        if not callable(fn):
            raise TypeError("Object is not callable")
        type_ = cls.callable_type(fn)
        io_details = cls.io_details(fn)
        return CallableDetails(
            args=io_details.args,
            has_kwargs=io_details.has_kwargs,
            return_type=io_details.return_type,
            is_routine=type_.is_routine,
            is_method=type_.is_method,
            is_async=type_.is_async,
            obj=fn,
        )

    @staticmethod
    def callable_type(fn: Any) -> CallableType:
        is_routine = inspect.isroutine(fn)
        return CallableType(
            is_routine=is_routine,
            is_method=inspect.ismethod(fn),
            is_async=inspect.iscoroutinefunction(fn if is_routine else fn.__call__),
        )

    @classmethod
    def io_details(cls, fn: Any) -> CallableIODetails:
        if not inspect.isroutine(fn):
            fn = fn.__call__
        signature = inspect.signature(fn)
        annotations = get_type_hints(fn)
        return CallableIODetails(
            args=[
                CallableArg(
                    name=param.name,
                    type=annotations.get(param.name),
                    is_positional=param.kind in cls._positional_kinds,
                )
                for param in signature.parameters.values()
                if param.kind in cls._kinds
            ],
            has_kwargs=any(
                param.kind == Parameter.VAR_KEYWORD
                for param in signature.parameters.values()
            ),
            return_type=annotations.get("return"),
        )



from typing import Any, Dict, List, Optional, Tuple

from mediator.event.base import EventPublisher


class EventAggregateError(Exception):
    pass


class ConfigEventAggregateError(AssertionError, EventAggregateError):
    pass


class EventAggregate:
    _publisher: Optional[EventPublisher]
    _staged: List[Tuple[Any, Dict[str, Any]]]

    def __init__(self):
        self._publisher = None
        self._staged = []

    def use(self, publisher: EventPublisher):
        self._publisher = publisher
        return self

    async def commit(self):
        publisher = self._publisher
        if publisher is None:
            raise ConfigEventAggregateError(f"Publisher is not set in {self!r}")

        async with publisher.transaction() as context:
            for obj, kwargs in self._staged:
                await context.publish(obj, **kwargs)

        self._staged.clear()

    def cleanup(self):
        self._staged.clear()

    def _enqueue(self, obj: Any, **kwargs):
        self._staged.append((obj, kwargs))



from typing import Any

from mediator.common.factory.base import HandlerFactory
from mediator.common.factory.policies import CallableHandlerPolicy, MethodHandlerPolicy
from mediator.common.factory.utils import (
    CallableAttributeDetails,
    CallableObjDetails,
    DirectHandlerCreate,
    HandlerSubjectArgGet,
)
from mediator.common.handler import Handler


class CallableHandlerFactory(HandlerFactory):
    def __init__(self, policy: CallableHandlerPolicy):
        self._obj_details = CallableObjDetails()
        self._arg_get = HandlerSubjectArgGet(name=policy.subject_arg)
        self._handler_create = DirectHandlerCreate(
            subject_as_keyword=policy.subject_as_keyword,
            arg_map=policy.arg_map,
            arg_strict=policy.arg_strict,
        )

    def create(self, obj: Any) -> Handler:
        details = self._obj_details(obj)
        arg = self._arg_get(details)
        return self._handler_create(details=details, arg=arg, obj=obj)


class MethodHandlerFactory(HandlerFactory):
    def __init__(self, policy: MethodHandlerPolicy):
        self._attribute_details = CallableAttributeDetails(
            name=policy.name, owner_subtype_of=policy.subtype_of
        )
        callable_policy = policy.policy
        self._arg_get = HandlerSubjectArgGet(name=callable_policy.subject_arg)
        self._handler_create = DirectHandlerCreate(
            subject_as_keyword=callable_policy.subject_as_keyword,
            arg_map=callable_policy.arg_map,
            arg_strict=callable_policy.arg_strict,
        )

    def create(self, obj: Any) -> Handler:
        details = self._attribute_details(obj)
        arg = self._arg_get(details)
        return self._handler_create(details=details, arg=arg, obj=obj)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mediator',
 'mediator.common',
 'mediator.common.factory',
 'mediator.common.handler',
 'mediator.common.registry',
 'mediator.event',
 'mediator.request',
 'mediator.utils']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'python-mediator',
    'version': '0.5.3',
    'description': 'Python mediator bus microframework for CQRS + ES',
    'long_description': '# python-mediator\n\n[![CI](https://github.com/dlski/python-mediator/workflows/CI/badge.svg?event=push)](https://github.com/dlski/python-mediator/actions?query=event%3Apush+branch%3Amaster+workflow%3ACI)\n[![codecov](https://codecov.io/gh/dlski/python-mediator/branch/master/graph/badge.svg?token=AU4T4Z81F6)](https://codecov.io/gh/dlski/python-mediator)\n[![pypi](https://img.shields.io/pypi/v/python-mediator.svg)](https://pypi.python.org/pypi/python-mediator)\n[![downloads](https://img.shields.io/pypi/dm/python-mediator.svg)](https://pypistats.org/packages/python-mediator)\n[![versions](https://img.shields.io/pypi/pyversions/python-mediator.svg)](https://github.com/dlski/python-mediator)\n[![license](https://img.shields.io/github/license/dlski/python-mediator.svg)](https://github.com/dlski/python-mediator/blob/master/LICENSE)\n\nElastic and extensible asyncio CQRS + ES python microframework.\nCompatible with recent python versions: 3.7, 3.8, 3.9, pypy3.\n\nCorresponds to clean architecture patterns, ideal for\ncommand/query segregation scenarios and event-driven design approaches.\nNo external dependencies - uses only standard libraries.\n\nKey features:\n- automatic function or method handler inspection -\n  proper action (command/query/event) to handler matching is fully automatic\n  and based on python type hints (annotations) by default\n- configurable middleware (operator) stack -\n  handler call flow can be extended easily\n  with i.e. data mapping, special exception handling or extra logging\n- configurable extra parameters injection\n- elastic and extensible -\n  custom behaviours and custom transport backends can be adapted with small effort\n\n## Help\nComing soon...\n\n## A command/query handling example\n```python\nfrom dataclasses import dataclass\n\nfrom mediator.request import LocalRequestBus\n\nbus = LocalRequestBus()\n\n\n@dataclass\nclass PrintMessageCommand:\n    message: str\n\n\n@bus.register\nasync def command_handler(event: PrintMessageCommand):\n    print(f"print message: {event.message}")\n    return event.message\n\n\n@dataclass\nclass DataQuery:\n    id: int\n\n\n@bus.register\nasync def query_handler(query: DataQuery):\n    print(f"data query: {query.id}")\n    return {"id": query.id, "data": "test"}\n\n\nasync def main():\n    printed_message = await bus.execute(PrintMessageCommand(message="test"))\n    assert printed_message == "test"\n\n    data = await bus.execute(DataQuery(id=1))\n    assert data == {"id": 1, "data": "test"}\n\n    # -- output --\n    # print message: test\n    # data query: 1\n\n```\nMore advanced example available in [tests/example/test_request_advanced.py](tests/example/test_request_advanced.py) for reference.\n\n## An event handling example\n```python\nfrom dataclasses import dataclass\n\nfrom mediator.event import LocalEventBus\n\nbus = LocalEventBus()\n\n\n@dataclass\nclass MessageEvent:\n    message: str\n\n\n@bus.register\nasync def first_handler(event: MessageEvent):\n    print(f"first handler: {event.message}")\n\n\n@bus.register\nasync def second_handler(event: MessageEvent):\n    print(f"second handler: {event.message}")\n\n\nasync def main():\n    await bus.publish(MessageEvent(message="test"))\n    # -- output --\n    # first handler: test\n    # second handler: test\n```\nMore advanced example available in [tests/example/test_event_advanced.py](tests/example/test_event_advanced.py) for reference.\n',
    'author': 'Damian Łukawski',
    'author_email': 'damian@lukawscy.pl',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/dlski/python-mediator',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from mediator.common.handler.base import Handler, HandlerInfo
from mediator.common.handler.handlers import DirectHandler, MappedDirectHandler

__all__ = [
    # base
    "Handler",
    "HandlerInfo",
    # handlers
    "DirectHandler",
    "MappedDirectHandler",
]



from typing import Sequence

from mediator.common.types import ActionCallType


class OperatorDef:
    def __eq__(self, other):
        raise NotImplementedError

    def __hash__(self):
        raise TypeError(f"{type(self)} is not hashable")

    def create(self, call: ActionCallType, **kwargs) -> ActionCallType:
        raise NotImplementedError


class OperatorStack:
    @classmethod
    def build(
        cls, op_defs: Sequence[OperatorDef], call: ActionCallType
    ) -> ActionCallType:
        target = call
        for op_def in reversed(op_defs):
            call = op_def.create(call, target=target)
        return call



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Type, Union


class MappablePolicy:
    def replace_map(self, policy: Any) -> "MappablePolicy":
        return self


@dataclass
class CallableHandlerPolicy(MappablePolicy):
    subject_arg: Optional[str] = None
    arg_map: Dict[str, str] = field(default_factory=dict)
    arg_strict: bool = False

    @property
    def subject_as_keyword(self):
        return bool(self.subject_arg)

    def replace_map(self, policy: Any) -> "CallableHandlerPolicy":
        if isinstance(policy, CallableHandlerPolicy):
            return policy
        else:
            return self


@dataclass
class MethodHandlerPolicy(MappablePolicy):
    name: str
    policy: CallableHandlerPolicy = field(default_factory=CallableHandlerPolicy)
    subtype_of: List[Type] = field(default_factory=list)

    @property
    def precises_type(self) -> bool:
        return bool(self.subtype_of)

    def replace_map(self, policy: Any) -> "MethodHandlerPolicy":
        if isinstance(policy, CallableHandlerPolicy):
            return MethodHandlerPolicy(
                name=self.name, policy=policy, subtype_of=self.subtype_of
            )
        elif isinstance(policy, MethodHandlerPolicy):
            return policy
        else:
            return self


PolicyType = Union[CallableHandlerPolicy, MethodHandlerPolicy]
